#include <stdlib.h>
#include <string.h>
#include "pbuffer.h"

#define NPO 50

typedef struct pnode
{
  struct pnode *next;
  point_t points[NPO];
  int     npoints;
} pnode_t;

struct pbuffer
{
  pnode_t *first, *last;
};

static pnode_t *pnode_new()
{
  return calloc(1,sizeof(pnode_t));
}

pbuffer_t *pbuffer_new()
{
  pbuffer_t *B;
  B= calloc(1,sizeof(*B));
  B->first= B->last= pnode_new();
  return B;
}

void pbuffer_free(pbuffer_t *B)
{
  pnode_t *P,*N;
  if (!B) return ;
  for(P=B->first;P;P=N)
  {
    N= P->next;
    free(P);
  }
  free(B);
}

void       pixel(pbuffer_t* B, int x,int y,color_t color)
{
  point_t po;
  po.x= x;
  po.y= y;
  po.color= color;

  if (B->last->npoints==NPO)
  {
    B->last->next= pnode_new();
    B->last= B->last->next;
  }
  B->last->points[B->last->npoints]= po;
  B->last->npoints++;
}


pbuffer_t* pbuffer_get(pbuffer_t *B, point_t **Rp, unsigned int *Rl)
{
  unsigned int np;
  pnode_t *P;
  point_t *D;
  np= 0;
  
  if (!B) { if (Rp) *Rp= NULL; if (Rl) *Rl= 0; return B; }
  for(P=B->first;P;P=P->next)
   np+= P->npoints;
  if (Rl) *Rl= np;
  if (Rp)
  {
    D= malloc(sizeof(point_t)*np);
    np= 0;
    for(P=B->first;P;P=P->next)
    {
      memcpy(D+np, P->points, sizeof(point_t)*P->npoints);
      np+= P->npoints;
    }
    *Rp= D;
  }
  return B;
}


void       pbuffer_clip(pbuffer_t *B,int x, int y, int w, int h)
{
  pnode_t *S;
  int i;
  if (!B) return ;
  for(S=B->first;S;S=S->next)
    for(i=0;i<S->npoints;i++)
      if (S->points[i].x<x ||
          S->points[i].y<y ||
          S->points[i].x>=(x+w) ||
          S->points[i].y>=(y+h))
      {
        S->points[i]= S->points[S->npoints-1];
        S->npoints--;
        i--;
      }
}

static int compoint(const void *pA, const void *pB)
{
  const point_t *A, *B;
  int R;
  A= pA; B= pB;
  R= A->x - B->x;
  if (R) return R;
  return A->y - B->y;
}

static void unify_block(pnode_t *P)
{
  int i,j;
  if (!P->npoints) return ;
  j= 0;
  for(i=1;i<P->npoints;i++)
    if (compoint(P->points+i, P->points+j))
    {
       j= j+1;
       if (i!=j) P->points[j]= P->points[i];
    }
  P->npoints= j+1;
}

static int find_point(point_t *P, int start, pnode_t *tab)
{
  int i,j,k,R;
  i= start;
  j= tab->npoints-1;
  while(i<=j)
  {
    k= (i+j)/2;
    R= compoint(P, tab->points+k);
    if (R>0) i= k+1;
    else if (R<0) j= k-1;
    else return k;
  }
  return -1;
}

static void unify_against(pnode_t *lop, pnode_t *fin)
{
  int nL, nF;
  int i,j,K,M;
  nL= lop->npoints;
  nF= fin->npoints;

  if (nL>nF) { unify_against(fin, lop); return ; }
  if (!nL || !nF) return ;

    // completely disjoint
  if (lop->points[nL-1].x<fin->points[0].x) return ;
  if (lop->points[0].x>fin->points[nF-1].x) return ;

  j= -1;
  K= 0;
  for(i=0;i<nL;i++)
  {
    M= find_point(lop->points+i, K, fin);
    if (M==-1)
    {
      j++;
      if (i!=j) lop->points[j]= lop->points[i];
    }
    else K= M;
  }
  lop->npoints= j+1;
}

void       pbuffer_unify(pbuffer_t *B)
{
  pnode_t *P, *R;
  if (!B || !B->first) return ;
  for(P=B->first;P;P=P->next)
  {
    qsort(P->points, P->npoints, sizeof(P->points[0]), compoint);
    unify_block(P);
  }
  for(P=B->first;P!=B->last;P=P->next)
   for(R=P->next;R;R=R->next)
     unify_against(P,R);
}

void       pbuffer_translate(pbuffer_t *B,int dx,int dy)
{
  int i;
  pnode_t *P;
  if (!B) return;
  for(P=B->first;P;P=P->next)
    for(i=0;i<P->npoints;i++)
    {
     P->points[i].x += dx;
     P->points[i].y += dy;
    }
}
