#include <stdio.h>
#include <stdlib.h>
#include <jpeglib.h>
#include <setjmp.h>

struct my_error_mgr {
  struct jpeg_error_mgr pub;	
  jmp_buf setjmp_buffer;	
};

typedef struct my_error_mgr * my_error_ptr;

static void my_error_exit (j_common_ptr cinfo)
{
  my_error_ptr myerr = (my_error_ptr) cinfo->err;
  (*cinfo->err->output_message) (cinfo);
  longjmp(myerr->setjmp_buffer, 1);
}

int write_jpeg(char *filename, int quality, unsigned int *D, int w, int h)
{
  struct jpeg_compress_struct cinfo;
  struct my_error_mgr jerr;
  FILE * outfile;		
  unsigned char *obuffer;
  JSAMPROW row_pointer[1];

  outfile= NULL;
  obuffer= NULL;

  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit = my_error_exit;

  if (setjmp(jerr.setjmp_buffer))
     goto error_compress;

  jpeg_create_compress(&cinfo);

  if ((outfile = fopen(filename, "wb")) == NULL)
  {
    fprintf(stderr, "can't open %s\n", filename);
    goto error_compress;
  }
  jpeg_stdio_dest(&cinfo, outfile);

  cinfo.image_width = w;
  cinfo.image_height = h;
  cinfo.input_components = 3;		
  cinfo.in_color_space = JCS_RGB; 	/* colorspace of input image */
  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo, quality, TRUE);

  jpeg_start_compress(&cinfo, TRUE);

  obuffer= malloc(3 * w);
  row_pointer[0]= (void*) obuffer;
  
  while (cinfo.next_scanline < cinfo.image_height)
  {
    int i;
    for(i=0;i<w;i++)
    {
      obuffer[3*i]= (D[i]>>16)&0xff;
      obuffer[3*i+1]= (D[i]>>8)&0xff;
      obuffer[3*i+2]= (D[i])&0xff;
    }
    (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
    D+= w;
  }

  jpeg_finish_compress(&cinfo);
  fclose(outfile);

  jpeg_destroy_compress(&cinfo);
  free(obuffer);
  return 0;
error_compress:
  if (outfile) fclose(outfile);
  if (obuffer) free(obuffer);
  jpeg_destroy_compress(&cinfo);
  return 1;
}

#ifdef SELF_TEST

#include <math.h>
#include <string.h>

#define W 300
#define H 300

int main()
{
  unsigned int D[W*H];
  int x,y;
  memset(D, 0, sizeof(D));
  for(x=0;x<W;x++)
    for(y=0;y<H;y++)
    {
      unsigned int P;
      double R;
      R=sqrt((x-W/2)*(x-W/2)+(y-H/2)*(y-H/2))/150.0; 
      if (R>1.0) R= 1.0;
      P= R* 255;
      D[y*W+x]= (P<<16) | (P<<8) | P;
    }
  if (write_jpeg("out.jpg", 90, D, W,H)) printf("failed\n");
  else  printf("out.jpg\n");
  return 0;
}
#endif
