#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include "list_dir.h"
#include "gfx.h"
#include "resize_image.h"
#include "readppm.h"
#include "pbuffer.h"
#include "line.h"
#include "rotate_image.h"
#include "wjpg.h"
#include "proc.h"
#include "cmdline.h"

static int imax2(int a,int b) { return a>b?a:b; }
static int imin2(int a,int b) { return a<b?a:b; }

enum { A_CULEF=1, A_CURIG, A_CUTOP, A_CUBOT , A_CUT, A_ROT1, A_ROT2 }; 
static int click_action;
static int move_action;

typedef struct 
{
  int is_src;
  unsigned int w,h,*data;
} img_t;

typedef struct input
{
  char *fn;
  img_t img;
} input_t;


static void linesel_end();

static int max_input_nr;
static int max_output_nr;
static int last_sfx_nr;

static int cc_l, cc_r, cc_t, cc_b;

  // display w, h and data
static unsigned int dw, dh, *dd;

static img_t *dimg;
static int di_x, di_y, di_w, di_h;
static input_t *dinput;

static char *scan_cmd;
static char *output_dir;
static char *scan_dir, *output_pfx;
static int scan_id, scan_fd[2];

static char scan_buf[2][2000];
static int scan_blen[2];


static int get_output(int fd, char *buf, int *len)
{
  int R;
  char c;
  R= read(fd, &c,1);
  if (R<0) return 1;
  if (R==0) return 0;
  if (c=='\r') return 0;
  if (c=='\n' || *len==1000) {  
      buf[*len]= 0;
      gfx_logview_printf(Wlog,"%s\n",buf); *len= 0; }
  else { buf[*len]= c; (*len)++; }
  return 0;
}

static int has_pfx(char *fn, char *pfx)
{
  int fL, pL;
  int i;
  fL= strlen(fn);
  pL= strlen(pfx);
  if (fL<=pL) return 0;
  for(i=0;i<pL;i++)
    if (toupper(pfx[i])!=toupper(fn[i])) return 0;
  return 1;
}

static int has_sfx(char *fn,char *sfx)
{
  int fL, sL;
  int i;
  fL= strlen(fn);
  sL= strlen(sfx);
  if (fL<=sL) return 0;
  fn+= fL-1;
  sfx+= sL-1; 
  for(i=0;i<sL;i++)
    if (toupper(sfx[-i])!=toupper(fn[-i])) return 0;
  return 1;
}

static int read_config()
{
  scan_cmd= getenv("CARTOSCAN_SCANCMD");
  output_dir= getenv("CARTOSCAN_OUTPUTDIR");
  scan_dir= getenv("CARTOSCAN_SCANDIR");
  output_pfx= getenv("CARTOSCAN_OUTPUTPFX");
  if (!scan_cmd || !output_dir || !scan_dir || !output_pfx) return 1;
  return 0;
}

static int next_scan_file()
{
  dir_t *D;
  int NR;
  diren_t *F;
  NR= 0;
  for(F=read_dir(D=open_dir(scan_dir,"DLPH"));F;F=read_dir(D))
  {
    if (!has_sfx(F->path,".ppm")) continue;
    if (has_pfx(F->name, "scan"))
       NR= imax2(NR, atoi(F->name+strlen("scan")));
  }
  close_dir(D);
  return NR+1;
}

static void scanimage()
{
  char **K;
  int err;
  int i;
  char out[2000];
  K= parse_cmd_line(scan_cmd,&err);
  if (!K) { gfx_logview_printf(Wlog, "bad scan command\n"); return ; }
  for(i=1;K[i];i++)
  {
    if (K[i][0]=='\'')
    {
     int j;
     K[i]++;
     for(j=0;K[i][j];j++) if (K[i][j]=='\'') { K[i][j]= 0; break; }
    }

    if (strchr(K[i],'@'))
    {
      char *z;
      strcpy(out,K[i]);
      z= strchr(out, '@');
      sprintf(z,"%s/scan%d.ppm",scan_dir,next_scan_file());
      K[i]= strdup(out);
    }
  }
  scan_id= proc_start(K[0],K,PROC_OUT | PROC_ERR | PROC_PTY);
  if (scan_id<0) return ;
  scan_fd[0]= proc_input(scan_id);
  scan_fd[1]= proc_error(scan_id);
  gfx_monitor_file(scan_fd[0], 0, 1,0,1);
  gfx_monitor_file(scan_fd[1], 1, 1,0,1);
  scan_blen[0]= scan_blen[1]= 0;
}

static int handle_scan_event(gfx_event_t* event)
{
  switch(event->type)
  {
  case GFXEV_READY_READ:
    if (get_output(scan_fd[event->key], scan_buf[event->key],
                   &scan_blen[event->key]) &&
        !proc_islive(scan_id)) goto finish;
    return 1;
  case GFXEV_READY_EXCEPT:
    gfx_stop_monitor_file(scan_fd[event->key], event->key);
    scan_fd[event->key]= -1;
    if (scan_fd[0]==scan_fd[1])
      goto finish;
    return 1;
  default:
    return 0;
  }
  
finish:
  proc_end(scan_id, NULL);
  scan_id= -1;
  if (scan_fd[0]>=0) gfx_stop_monitor_file(scan_fd[0], 0);
  if (scan_fd[1]>=0) gfx_stop_monitor_file(scan_fd[1], 1);
  scan_fd[0]= scan_fd[1]= -1;
  gfx_filelist_refresh(Wfiles);
  return 1;
}

static void list_output_dir()
{
  dir_t *D;
  diren_t *F;
  for(F=read_dir(D=open_dir(output_dir,"DLPH"));F;F=read_dir(D))
  {
    if (!has_sfx(F->path,".jpg")) continue;
    if (has_pfx(F->name, output_pfx))
       max_output_nr= imax2(max_output_nr, atoi(F->name+strlen(output_pfx)));
  }
  close_dir(D);
  max_output_nr++;
}

static void make_dirs()
{
  mkdir(scan_dir, 0755);
  mkdir(output_dir, 0755);
  list_output_dir();
}

static void replace_image(img_t *img)
{
  unsigned int w, h, *d,x,y,ox,oy;
  if (dimg && !dimg->is_src) { free(dimg->data); free(dimg); }
  dimg= img;
  for(y=0;y<dh;y++) for(x=0;x<dw;x++) 
  {
    dd[x+y*dw]= (x+y)%7 > 3 ? 0xffffff : 0xcccccc;
  }
  if (!img) goto done;
  if (img->w<=dw && img->h<=dh)
  {
    w= img->w; h= img->h; d= img->data;
  }
  else
  {
    resize_image_argb(img->data,img->w,img->h, dw, dh, &d,&w,&h);
  }
  ox= (dw-w)/2;
  oy= (dh-h)/2;
  for(y=0;y<h;y++)
   for(x=0;x<w;x++)
     dd[ (x+ox)+(y+oy)*dw ]= d[x+y*w];
  if (d!=img->data) free(d);
  di_x= ox;
  di_y= oy;
  di_w= w;
  di_h= h;
done:
  gfx_canvas_update(Wcanvas,0,0,dw,dh);
}

static void save_file()
{
  char buf[2000];
  int x,y;
  img_t *ii;
  if (!dimg) return ;
  sprintf(buf,"%s/%s%d-%d.jpg",output_dir, output_pfx, max_output_nr, 
                            last_sfx_nr);
  last_sfx_nr++;
  gfx_logview_printf(Wlog,"writing %s\n",buf);
  write_jpeg(buf,100,dimg->data,dimg->w,dimg->h);
  
  ii= &dinput->img;
  for(y=cc_t;y+cc_b<ii->h;y++)
   for(x=cc_l;x+cc_r<ii->w;x++)
     ii->data[x+y*ii->w]= 0x00;
  replace_image(ii);
  cc_l= cc_r= cc_t= cc_b= 0;
}

static void write_ppm_file(input_t *inp)
{
  FILE *f;
  f= fopen(inp->fn, "wb");
  if (!f) 
  { 
    gfx_logview_printf(Wlog,"can't write %s: %s\n",inp->fn, strerror(errno));
    return ;
  }
  write_ppm(f, inp->img.data, inp->img.w, inp->img.h);
  fclose(f);
}


static void load_file(char *fn)
{
  FILE *f;
  int w,h,*data;

  if (dinput && !strcmp(dinput->fn, fn)) { free(fn); return ; }
  if (dimg && dinput && dimg==&dinput->img) dimg= NULL;
  if (dinput)
  { 
    write_ppm_file(dinput); 
    free(dinput->fn);
    free(dinput->img.data);
    free(dinput);
    dinput= NULL;
  }
  if (dimg) { free(dimg->data); free(dimg); dimg= NULL; }
  
  f= fopen(fn,"rb");
  if (!f)
  {
    gfx_logview_printf(Wlog,"can't open %s: %s\n",fn, strerror(errno));
    replace_image(NULL);
    return ;
  }
  read_ppm(f, &w,&h,&data);
  fclose(f);
  dinput= calloc(1,sizeof(*dinput));
  dinput->img.data= (unsigned int*) data;
  dinput->img.w= w;
  dinput->img.h= h;
  dinput->img.is_src= 1;
  dinput->fn= fn;
  replace_image(&dinput->img);
  gfx_logview_printf(Wlog,"loaded %s\n", fn);
  cc_l= cc_r= cc_t= cc_b= 0;
}

static void replace_input()
{
  if (!dimg || dimg->is_src) return ;
  if (!dinput) return ;
  if (dinput->img.data) free(dinput->img.data);
  dinput->img.data= dimg->data;
  dinput->img.w= dimg->w;
  dinput->img.h= dimg->h;
  dinput->img.is_src= 1;
  free(dimg);
  dimg= &dinput->img;
  cc_l= cc_r= cc_t= cc_b= 0;
}


static int hilight_x, hilight_y, hilight_type;
static point_t* hilight_pts;
static int hilight_npts;
static int hilighted;
static int hilight_active;




static void hilight()
{
  int minx, miny, maxx, maxy;
  int x,y,ox,oy,i;
  if (!hilight_pts) return ;
  if (!hilight_type) return ;
  if (hilight_type=='X') { ox= hilight_x; oy= di_y; }
                    else { ox= di_x; oy= hilight_y; }

  minx= maxx= hilight_pts[0].x+ox;
  miny= maxy= hilight_pts[0].y+oy;
  for(i=0;i<hilight_npts;i++)
   {
     x= ox+hilight_pts[i].x;
     y= oy+hilight_pts[i].y;
     dd[x+y*dw] ^= 0xffFFff;
     minx= imin2(minx, x); miny= imin2(miny, y);
     maxx= imax2(maxx, x); maxy= imax2(maxy, y);
   }

  gfx_canvas_update(Wcanvas, minx, miny, maxx-minx+1, maxy-miny+1);
}

static void start_hilight(int type, pbuffer_t *obj)
{
  linesel_end();
  if (hilighted) hilight();
  if (hilight_pts) free(hilight_pts);
  pbuffer_get(obj,&hilight_pts,&hilight_npts);
  pbuffer_free(obj);
  hilighted= 0;
  hilight_type= type;
  gfx_canvas_begin_report_mouse_position(Wcanvas);
  hilight_active= 1;
}

static void hilight_pos(int x,int y)
{
  if (hilighted) hilight();
  hilighted= 1;
  hilight_x= x;
  hilight_y= y;
  hilight();
}

static void end_hilight()
{
  if (!hilight_active) return ;
  if (hilighted)
    hilight();
  if (hilight_pts) free(hilight_pts);
  hilight_pts= NULL;
  hilight_type= 0;
  hilighted= 0;
  gfx_canvas_end_report_mouse_position(Wcanvas);
  hilight_active= 0;
}

static pbuffer_t *horz_line()
{
  pbuffer_t *B;
  int x;
  B= pbuffer_new();
  for(x=0;x<di_w;x++)
   pixel(B, x, 0, 0xff0000);
  return B;
}

static pbuffer_t *vert_line()
{
  pbuffer_t *B;
  int y;
  B= pbuffer_new();
  for(y=0;y<di_h;y++)
   pixel(B, 0, y, 0xff0000);
  return B;
}

static void do_cut(int left, int top, int right, int bot)
{
  img_t *K;
  int ox, oy, cw, ch, x, y;
  if (!dimg || !di_w || !di_h) return ;
  left= imin2(imax2(0, left), di_w) * dimg->w / di_w;
  right= imin2(imax2(0, right), di_w) * dimg->w / di_w;
  top= imin2(imax2(0, top), di_h) * dimg->h / di_h;
  bot= imin2(imax2(0, bot), di_h) * dimg->h / di_h;
  if (!left && !right && !top && !bot) return ;

  ox= left;
  oy= top;
  cw= dimg->w - left - right;
  ch= dimg->h - top - bot;
  if (cw<=0 || ch<=0 || ox>=dimg->w || oy>=dimg->h) return ;

  K= malloc(sizeof(*K));
  K->w= cw; 
  K->h= ch;
  K->is_src= 0;
  K->data= malloc(sizeof(K->data[0])*cw*ch);
  for(y=0;y<ch;y++)
    for(x=0;x<cw;x++)
      K->data[x+y*cw]= dimg->data[(x+ox)+(y+oy)*dimg->w];
  replace_image(K);

  cc_l+= left; cc_r+= right; cc_t+= top; cc_b+= bot;
}

static int linesel_x0, linesel_y0, linesel_x1, linesel_y1;
static point_t *linesel_pts;
static unsigned int linesel_npts;

static void draw_linesel_obj(int f)
{
  int i,x,y;
  int minx,maxx,miny,maxy;
  if (!linesel_pts) return ;
  minx= maxx= linesel_pts[0].x;
  miny= maxy= linesel_pts[0].y;
  for(i=0;i<linesel_npts;i++)
  {
    x= linesel_pts[i].x;
    y= linesel_pts[i].y;
    if (x<0 || x>=dw || y<=0 || y>=dh) continue;
    dd[x+y*dw] ^= 0xffffff;
    minx= imin2(minx, x); maxx= imax2(maxx, x);
    miny= imin2(miny, y); maxy= imax2(maxy, y);
  }
  gfx_canvas_update(Wcanvas, minx, miny, maxx-minx+1, maxy-miny+1);
  if (!f) return ;
  free(linesel_pts);
  linesel_pts= NULL;
}

static void linesel_begin(int x,int y)
{
  end_hilight();
  linesel_x0= x;
  linesel_y0= y;
  linesel_pts= NULL;
  linesel_npts= 0;
  gfx_canvas_begin_report_mouse_position(Wcanvas);
}

static void linesel(int x,int y)
{
  pbuffer_t *B;
  linesel_x1= x;
  linesel_y1= y;

  if (linesel_pts)  draw_linesel_obj(1); 
  B= line(NULL,0xff0000,linesel_x0,linesel_y0,linesel_x1,linesel_y1);
  pbuffer_get(B, &linesel_pts, &linesel_npts);
  pbuffer_free(B);
  draw_linesel_obj(0);
}

static void linesel_end()
{
  if (!linesel_pts) return ;
  draw_linesel_obj(1);
  gfx_canvas_end_report_mouse_position(Wcanvas);
}

static void do_rotate()
{
  double angle;
  img_t *ii;
  int x0,x1,y0,y1;
  linesel_end(); 
  gfx_logview_printf(Wlog, "linesel [%d,%d]->[%d,%d]\n",
              linesel_x0,linesel_y0, linesel_x1,linesel_y1);
  if (linesel_x0>linesel_x1) { x0= linesel_x1; x1= linesel_x0; 
               y0= linesel_y1; y1= linesel_y0; }
        else { x0= linesel_x0; x1= linesel_x1; 
               y0= linesel_y0; y1= linesel_y1; }
  angle= atan2(y1-y0, x1-x0) * 180/M_PI;
  gfx_logview_printf(Wlog, "rotate %d degrees\n", (int) angle);
  if (!dimg) return ;
  ii= malloc(sizeof(*ii));
  ii->is_src= 0;
  rotate_image_double(dimg->data, dimg->w, dimg->h, -angle,
                      &ii->data, &ii->w, &ii->h);
  replace_image(ii);
}

static void handle_click(int ex,int ey)
{
  int cu_l, cu_t, cu_b, cu_r, x,y;
  cu_l= cu_t= cu_b= cu_r= 0;
  x= hilight_x - di_x;
  y= hilight_y - di_y;

  switch(click_action)
  {
  case A_CUTOP: cu_t= y; break;
  case A_CUBOT: cu_b= di_h - y; break;
  case A_CULEF: cu_l= x; break; 
  case A_CURIG: cu_r= di_w - x; break;
  case A_ROT1: linesel_begin(ex, ey); click_action= A_ROT2; break;
  case A_ROT2: do_rotate(); click_action= 0; break;
  }

  switch(click_action)
  {
  case A_CUTOP: 
  case A_CUBOT:
  case A_CULEF:
  case A_CURIG:
    end_hilight();
    do_cut(cu_l,cu_t,cu_r,cu_b);
    click_action= 0;
    break;
  }
}

static img_t *rotate_begin()
{
  img_t *ii;
  end_hilight();
  linesel_end();
  if (!dimg) return NULL;
  ii= malloc(sizeof(*ii));
  ii->w= dimg->h;
  ii->h= dimg->w;
  ii->data= malloc(sizeof(ii->data[0])*ii->w*ii->h);
  ii->is_src= 0;
  return ii;
}

static void rotate_left()
{
  int x,y,xp,yp;
  img_t *ii;
  if (!(ii=rotate_begin())) return ;
  for(y=0;y<dimg->h;y++)
  {
    xp= y;
    for(x=0;x<dimg->w;x++)
    {
      yp= dimg->w-1-x;
      ii->data[xp+yp*ii->w]= dimg->data[x+y*dimg->w];
    }
  }
  replace_image(ii);
}

static void rotate_right()
{
  int x,y,xp,yp;
  img_t *ii;
  if (!(ii=rotate_begin())) return ;
  for(y=0;y<dimg->h;y++)
  {
    xp= dimg->h-1-y;
    for(x=0;x<dimg->w;x++)
    {
      yp= x;
      ii->data[xp+yp*ii->w]= dimg->data[x+y*dimg->w];
    }
  }
  replace_image(ii);
}

static void delfile()
{
  if (!dinput) return ;
  if (unlink(dinput->fn))
  {
    gfx_logview_printf(Wlog,"can't remove %s: %s\n",
        dinput->fn, strerror(errno));
    return ;
  } 
  gfx_logview_printf(Wlog,"deleted %s\n",dinput->fn);
  if (dimg && dimg->is_src) dimg= NULL;
  free(dinput->img.data);
  free(dinput->fn);
  free(dinput);
  dinput= NULL;
  replace_image(NULL);
  gfx_filelist_refresh(Wfiles);
}

static void cut_auto()
{
  int t,l,r,b,x,y;
  unsigned int P,*d;
  int w,h;
  if (!dimg) return ;
  d= dimg->data;
  P= d[0];
  w= dimg->w;
  h= dimg->h;
  for(t=0;t<h;t++) for(x=0;x<w;x++) if (d[x+t*w]!=P) goto f1;
f1:
  for(b=0;b<h;b++) for(x=0;x<w;x++) if (d[x+(h-1-b)*w]!=P) goto f2;
f2:
  for(l=0;l<w;l++) for(y=0;y<h;y++) if (d[l+y*w]!=P) goto f3;
f3:
  for(r=0;r<w;r++) for(y=0;y<h;y++) if (d[(w-1-r)+y*w]!=P) goto f4;
f4:
  do_cut(l,t,r,b);
}

int main()
{
  static char *ppmfiles[]= { "PPM images (*.ppm *.pnm)", NULL };
  gfx_init(NULL, 1000, 800, "Cartoon Scan");

  gfx_SCREEN_0();

  if (read_config())
  {
    gfx_ask("Bad configuration","!OK", "Please set the environment variables\n"
     "   CARTOSCAN_SCANCMD\n"
     "   CARTOSCAN_SCANDIR\n"
     "   CARTOSCAN_OUTPUTDIR\n"
     "   CARTOSCAN_OUTPUTPFX\n"
     "for this program to work properly.");
    goto done;
  }

  gfx_canvas_get_buffer(Wcanvas,&dw,&dh,&dd);

  make_dirs();

  gfx_filelist_set_directory(Wfiles, scan_dir);
  gfx_filelist_set_types(Wfiles, ppmfiles);

  replace_image(NULL);

  while(1)
  {
   gfx_event_t ev;
   if (!gfx_event(NULL,&ev)) continue;
   if (handle_scan_event(&ev)) continue;
   switch(ev.type)
   {
   case Wcutop: 
     gfx_logview_printf(Wlog,"click for top cut\n");
     start_hilight('Y', horz_line());
     click_action= A_CUTOP; move_action= A_CUT;
     break;
   case Wcubot: 
     gfx_logview_printf(Wlog,"click for bottom cut\n");
     start_hilight('Y', horz_line());
     click_action= A_CUBOT; move_action= A_CUT;
     break;
   case Wculeft:
     gfx_logview_printf(Wlog,"click for left cut\n");
     start_hilight('X', vert_line());
     click_action= A_CULEF; move_action= A_CUT;
     break;
   case Wcuright:
     gfx_logview_printf(Wlog,"click for right cut\n");
     start_hilight('X', vert_line());
     click_action= A_CURIG; move_action= A_CUT;
     break;
   case Wrosel: click_action= A_ROT1; move_action= A_ROT1; break;
   case GFXEV(Wcanvas,POINTER_MOVE):
     if (move_action==A_CUT) hilight_pos(ev.x,ev.y);
                       else  linesel(ev.x,ev.y);
     break;
   case GFXEV(Wcanvas,BUTTON_PRESS): handle_click(ev.x,ev.y); break;
   case Wroleft: rotate_left(); break;
   case Wroright: rotate_right(); break;
   case Wsave: save_file(); break;
   case Wup: if (dinput) replace_image(&dinput->img); break;
   case Wreplace:  replace_input(); break;
   case Wfiles: load_file(ev.string); break;
   case GFXEV(Wfiles,DOUBLE_CLICK): free(ev.string); break;
   case GFXEV_QUIT: if (dinput) write_ppm_file(dinput); goto done;
   case Wscan: scanimage(); break;
   case Wdel: delfile(); break;
   case Wcuauto: cut_auto(); break;
   }
  }

done:
  gfx_close();
  return 0;
}
